# frozen_string_literal: true

module Hyrax
  module Crc1280Experiments
    class ComplexSessionsController < ApplicationController
      include HyraxHelper
      include ComplexHelper
      before_action :set_crc1280_experiment
      before_action :set_complex_subject, only: %i[new create validate_session_title]
      before_action :set_complex_session, only: %i[show edit update]
      before_action :check_tombstone, except: %i[validate_session_title]
      before_action :authorize
      before_action :sanitise_params, only: [:create, :update]
      with_themed_layout :decide_layout

      def new
        @complex_session = ::ComplexSession.new
      end

      def edit
        @complex_subject = @complex_session.complex_subject
      end

      def show
        @complex_subject = @complex_session.complex_subject
        @session_members = Kaminari.paginate_array(@complex_session.file_sets + ComplexModality.sort_by_title(@complex_session.source_identifier)).page(params[:page] || 1).per(params[:per_page] || 10)
        @workflow = WorkflowPresenter.new(@crc1280_experiment, current_ability)
      end

      def create
        @complex_session = ::ComplexSession.new(complex_session_params.merge!({ parent_source_identifier: @complex_subject.source_identifier }))

        if @complex_session.save
          update_file_set
          message = 'Complex session was successfully created.'
          flash[:alert] = save_alert if params[:uploaded_files].present?
          redirect_to crc1280_experiment_complex_session_path(@complex_session.crc1280_experiment, @complex_session), notice: message
        else
          render :new
        end
      end

      def update
        if @complex_session.update(complex_session_params)
          update_file_set
          message = 'Complex session was successfully updated.'
          flash[:alert] = save_alert if params[:uploaded_files].present?
          redirect_to crc1280_experiment_complex_session_path(@complex_session.crc1280_experiment, @complex_session), notice: message
        else
          @complex_subject = @complex_session.complex_subject
          render :edit
        end
      end

      def validate_session_title
        folder_name = sanitize_title(params[:folder_name])
        sessions = @complex_subject.complex_sessions
        .where(s3_folder_name: folder_name)
        .where.not(id: params[:current_object_id])

        invalid_length = false
        if folder_name.length < 3 or folder_name.length > 63
          invalid_length = true
          message = I18n.t('rdms.errors.complex_session.title.length', value: folder_name)
        end
        message = I18n.t('rdms.errors.complex_session.title.uniqueness', value: folder_name) if sessions.present?
        render json: { data_exist: (sessions.present? || invalid_length), error_message: message }
      end

      private

      def set_crc1280_experiment
        @crc1280_experiment = ::Hyrax.query_service.find_by(id: params[:crc1280_experiment_id])
      end

      def set_complex_subject
        @complex_subject = ComplexSubject.find(params[:add_sessions_to_subject] || params[:parent_id])
      end

      def set_complex_session
        @complex_session = ::ComplexSession.find(params[:id])
      end

      def complex_session_params
        params.require(:complex_session).permit(:session_title, :session_date_recorded)
      end

      def decide_layout
        layout = case action_name
                 when 'show'
                   '1_column'
                 else
                   'dashboard'
                 end
        File.join(theme, layout)
      end

      def update_file_set
        if params[:uploaded_files].present? and params[:uploaded_files].any?
          uploaded_files = Hyrax::UploadedFile.find(params[:uploaded_files])
          Hyrax::WorkUploadsHandler.new(work: @crc1280_experiment).add(files: uploaded_files, file_set_params: [{for_complex_identifier: @complex_session.source_identifier, for_complex_type: 'ComplexSession'}]).attach
        end
      end

      def save_alert
        'Your files are being processed by RDMS in the background. The metadata and access controls you specified are being applied. You may need to refresh this page to see these updates.'
      end

      def check_tombstone
        if @crc1280_experiment.is_tombstoned == true
          redirect_to main_app.polymorphic_path(@crc1280_experiment)
        end
      end

      def sanitise_params
        params[:complex_session][:session_date_recorded] = DateTime.strptime(params[:complex_session][:session_date_recorded], "%m/%d/%Y") if params[:complex_session][:session_date_recorded].present?
      end
    end
  end
end
