# app/models/concerns/file_relocation_concern.rb
require 'json'
module S3FileHandleable
  extend ActiveSupport::Concern
  include S3FileHandler

  def relocate_files_on_folder_name_change
    return true unless saved_change_to_s3_folder_name?
    name_changes = saved_changes[:s3_folder_name]
    return true if name_changes[0].nil?

    Hyrax::HandleS3FolderNameJob.perform_later(self.class.name, self.id, name_changes)
  end

  def clear_s3_and_all_associated_objects
    s3 = S3StorageService.new
    s3.init_client
    if self.class.in?([Dataset, Crc1280Experiment])
      bucket_name = s3.sanitise_name(ENV.fetch("S3_FILE_UPLOAD_BUCKET"))
      user = ::User.find_by_user_key(self.depositor)
      all_objects, size = s3.list_all_objects(bucket_name, self.id.to_s)
      s3.delete_objects(bucket_name, all_objects.map { |obj| obj[:key] }) if all_objects.present?
      if self.class == Crc1280Experiment
        self.complex_subjects.destroy_all
      end
    else
      bucket_name =  s3.sanitise_name(ENV.fetch("S3_FILE_UPLOAD_BUCKET"))
      user = ::User.find_by_user_key(self.crc1280_experiment.depositor)
      objects_prefix = objects_prefix(self, s3_folder_name)
      all_objects, size = s3.list_all_objects(bucket_name, objects_prefix)
      s3.delete_objects(bucket_name, all_objects.map{|obj| obj[:key]}) if all_objects.present?
    end

    self.file_sets.each do |file_set|
      transactions = Hyrax::Transactions::Container
      transactions['file_set.destroy']
      .with_step_args('file_set.remove_from_work' => { user: user },
                      'file_set.delete' => { user: user })
      .call(file_set)
      .value!
    end
  end

  def save_meta_json_in_s3
    s3 = S3StorageService.new
    s3.init_client

    crc1280_experiment = self.is_a?(Crc1280Experiment) ? self : self.crc1280_experiment
    bucket_name = s3.sanitise_name(ENV.fetch("S3_FILE_UPLOAD_BUCKET"))
    complex_modalities = self.is_a?(ComplexModality) ? [self] : self.complex_modalities.to_a.flatten

    complex_modalities.each do |modality|
      subject = modality.complex_subject
      session = modality.complex_session
      
      meta_file_key = "#{crc1280_experiment.id.to_s}/#{subject.s3_folder_name}/#{session.s3_folder_name}/#{modality.s3_folder_name}/meta.json"
      json_data = prepare_json(crc1280_experiment, subject, session, modality)

      s3.add_content(bucket_name, meta_file_key, json_data)
    end
  end

  private

  def update_crc1280_experiment_date_modified_and_files_data
    return true unless crc1280_experiment.present?
    
    s3service = S3StorageService.new
    s3service.init_client
    bucket_name = s3service.sanitise_name(ENV.fetch("S3_FILE_UPLOAD_BUCKET"))

    list_of_objects, total_size, list_of_format = s3service.list_all_objects(bucket_name, nil, true)

    crc1280_experiment.date_modified = DateTime.now
    crc1280_experiment.files_size = [total_size.to_s]
    crc1280_experiment.format = list_of_format

    Hyrax.persister.save(resource: crc1280_experiment)
  end


  def prepare_json(crc1280_experiment, subject, session, modality)
    @meta = {}
    data_from_crc1280_experiment(crc1280_experiment)
    data_from_complex_subject(subject)
    data_from_complex_session(session)
    data_from_complex_modality(modality)
    JSON.pretty_generate(@meta)
  end

  def data_from_complex_subject(subject)
    @meta["Subject ID"] = ""
    @meta["Subject ID"] = subject.subject_title.strip if subject.subject_title.present?

    @meta["Subject Species"] = ""
    @meta["Subject Species"] = subject.subject_species.strip if subject.subject_species.present?

    @meta["Subject type"] = ""
    @meta["Subject type"] = subject.subject_type.strip if subject.subject_type.present?

    @meta["Subject sex"] = ""
    @meta["Subject sex"] = subject.subject_sex.strip if subject.subject_sex.present?

    @meta["Subject age"] = ""
    @meta["Subject age"] = subject.subject_age.to_s.strip if subject.subject_age.present?
  end

  def data_from_complex_session(session)
    # date from complex session will overwrite date from complex date
    if session.present? and session.session_date_recorded.present?
      date_recorded = session.session_date_recorded.strftime("%Y-%m-%d")
      @meta["Record date"] =  date_recorded
      @meta["DataCite-Date"] = date_recorded
      @meta["DataCite-dateType"] = "Created"
      @meta["DublinCore-Date"] = date_recorded
    end
  end

  def data_from_complex_modality(modality)
    @meta["Modality"] = []
    @meta["Modality"] << modality.modality if modality.present? and modality.modality.present?
    @meta["DataCite-ResourceType"] = @meta["Modality"]
  end

  def data_from_crc1280_experiment(crc1280_experiment)
    @meta["DublinCore-Title"] = crc1280_experiment.title[0]
    @meta["Experiment title"] = crc1280_experiment.title[0]
    @meta["DataCite-Title"] = crc1280_experiment.title[0]
    @meta["DataCite-titleType"] = ""
    @meta["DataCite-resourceTypeGeneral"] = "Collection"
    @meta["DublinCore-Type"] = "Collection"
    @meta["DublinCore-Source"] = ""
    

    # Language
    @meta["DublinCore-Language"] = ""
    @meta["DataCite-Language"] = ""
    if crc1280_experiment.language.any? and crc1280_experiment.language.first.present?
      @meta["DublinCore-Language"] = crc1280_experiment.language[0].strip
      @meta["DataCite-Language"] = crc1280_experiment.language[0].strip
    end

    # Size and Format
    @meta["DataCite-Size"] = crc1280_experiment.files_size.join("; ")
    @meta["DataCite-Format"] = crc1280_experiment.format.join("; ")
    @meta["DublinCore-Format"] = crc1280_experiment.format.join("; ")

    if crc1280_experiment.crc_resource_type.present?
      @meta["Resource Type"] = crc1280_experiment.crc_resource_type.strip
    end

    @meta["DataCite-Description"] = ""
    @meta["DublinCore-Description"] = ""
    @meta["DataCite-descriptionType"] = "Abstract"
    if crc1280_experiment.experiment_description.any? and crc1280_experiment.experiment_description.first.present?
      experiment_description = crc1280_experiment.experiment_description[0].strip
      @meta["DataCite-Description"] = experiment_description
      @meta["DublinCore-Description"] = experiment_description
      @meta["Experiment Description"] = experiment_description
    end

    @meta["Group ID"] = ""
    if crc1280_experiment.parent_collections.any? and
       crc1280_experiment.parent_collections.first.present? and
       crc1280_experiment.parent_collections.first.title.present?
      @meta["Group ID"] = crc1280_experiment.parent_collections.first.title[0].strip
    end
    @meta["Shared With"] = ""

    @meta["Animal|Ethics approval No."] = ""
    if crc1280_experiment.approval_number.present?
      @meta["Animal|Ethics approval No."] = crc1280_experiment.approval_number.to_s.strip
    end

    # Keywords/Subject
    @meta["DataCite-subjectSchemeURI"] = ""
    @meta["DataCite-subjectScheme"] = ""
    @meta["DataCite-subjectValueURI"] = ""
    @meta["DublinCore-Subject"] = ""
    @meta["DataCite-Subject"] = ""

    if crc1280_experiment.keyword.any?
      keywords = crc1280_experiment.keyword.reject(&:empty?).map { |s| s.strip }
      @meta["DublinCore-Subject"] = keywords.join("; ") if keywords.any?
      @meta["DataCite-Subject"] = keywords.join("; ") if keywords.any?
    end

    #Publisher
    @meta["DublinCore-Publisher"] = ""
    @meta["DataCite-Publisher"] = ""
    if crc1280_experiment.publisher.any? and crc1280_experiment.publisher.first.present?
      @meta["DublinCore-Publisher"] =  crc1280_experiment.publisher[0].strip
      @meta["DataCite-Publisher"] =  crc1280_experiment.publisher[0].strip
    end

    @meta["DublinCore-Coverage"] = ""
    if crc1280_experiment.coverage.present?
      coverage = crc1280_experiment.coverage.strip
      @meta["DublinCore-Coverage"] = coverage
      @meta["DataCite-GeoLocationPlace"] = coverage
    end

    @meta["Software version"] = crc1280_experiment.software_version.any? ? crc1280_experiment.software_version.first : ""
    @meta["DataCite-Version"] = "4.4"
    @meta["Extra information"] = ""
    if crc1280_experiment.extra_information.any? and crc1280_experiment.extra_information.first.present?
      @meta["Extra information"] = crc1280_experiment.extra_information[0].strip
    end

    # Rights
    @meta["DataCite-Rights"] = ""
    @meta["DublinCore-Rights"] = ""
    @meta["DataCite-rightsURI"] = ""
    @meta["DataCite-rightsIdentifierScheme"] = ""
    @meta["DataCite-rightsSchemeURI"] = ""
    @meta["DataCite-rightsIdentifier"] = ""

    if crc1280_experiment.license.any? and crc1280_experiment.license.first.present?
      rights = LicenseService.new.label(crc1280_experiment.license.first)
      @meta["DataCite-Rights"] = rights
      @meta["DublinCore-Rights"] = rights
      @meta["DataCite-rightsURI"] = crc1280_experiment.license.first
    end

    # Map ARK and DOI
    @meta["DataCite-alternateIdentifier"] = ""
    @meta["DataCite-alternateIdentifierType"] = ""
    @meta["DataCite-Identifier"] = ""
    @meta["DublinCore-Identifier"] = ""
    @meta["DataCite-identifierType"] = ""
    ark = ''
    
    crc1280_experiment.complex_identifier.each do |idf|
      scheme = idf.fetch(:scheme, '')
      identifier = idf.fetch(:identifier, '')
      if identifier == "ARK" and identifier.present?
        ark = identifier.strip
        @meta["DataCite-alternateIdentifier"] = ark
        @meta["DataCite-alternateIdentifierType"] = "ARK"
      elsif scheme == "DOI" and identifier.present?
        @meta["DublinCore-Identifier"] = identifier.strip
        @meta["DataCite-Identifier"] = identifier.strip
        @meta["DataCite-identifierType"] = "DOI"
      end
    end
      
    if ark.empty?
      @meta["DataCite-alternateIdentifier"] = crc1280_experiment.id.to_s
      @meta["DataCite-alternateIdentifierType"] = "Local"
    end
    
    data_from_complex_person(crc1280_experiment.complex_person)
    data_from_complex_date(crc1280_experiment.complex_date)
    data_from_complex_funding_reference(crc1280_experiment.complex_funding_reference)
    data_from_complex_relation(crc1280_experiment.complex_relation)
  end

  def data_from_complex_person(complex_person)
    complex_creators, complex_contributors = group_people_by_role(complex_person)
    # Creators
    creators = get_people_info(complex_creators)
    # names, first_names, last_names, affiliations, name_identifiers, roles
    @meta["Creator"] = creators[:names]
    @meta["DublinCore-Creator"] = creators[:names]
    @meta["DataCite-creatorName"] = creators[:names]
    @meta["DataCite-creatorNameType"] = creators[:name_types]
    @meta["DataCite-creatorFamilyName"] = creators[:last_names]
    @meta["DataCite-creatorGivenName"] = creators[:first_names]
    @meta["DataCite-creatorAffiliation"] = creators[:affiliations]
    @meta["DataCite-creatorNameIdentifier"] = creators[:name_identifiers]
    @meta["DataCite-creatorNameIdentifierScheme"] = creators[:name_identifier_schemes]
    @meta["DataCite-creatorSchemeURI"] = creators[:scheme_uri]

    # Contributors
    contributors = get_people_info(complex_contributors)
    # names, first_names, last_names, affiliations, name_identifiers, roles
    @meta["Contributor"] = contributors[:names]
    @meta["DublinCore-Contributor"] = contributors[:names]
    @meta["DataCite-contributorName"] = contributors[:names]
    @meta["DataCite-contributorNameType"] = contributors[:name_types]
    @meta["DataCite-contributorFamilyName"] = contributors[:last_names]
    @meta["DataCite-contributorGivenName"] = contributors[:first_names]
    @meta["DataCite-contirbutorAffiliation"] = contributors[:affiliations]
    @meta["DataCite-contributorType"] = contributors[:roles]
    @meta["DataCite-contributorNameIdentifier"] = contributors[:name_identifiers]
    @meta["DataCite-contributorNameIdentifierScheme"] = contributors[:name_identifier_schemes]
    @meta["DataCite-contributorSchemeURI"] = contributors[:scheme_uri]
  end

  def group_people_by_role(complex_person)
    creators = []
    contributors = []
    complex_person.each do |person|
      role = person.fetch(:role, '').strip if person.fetch(:role, '').present?
      if role == "creator"
        creators << person
      else
        contributors << person
      end
    end
    return creators, contributors
  end

  def get_people_info(complex_people)
    names = []
    first_names = []
    last_names = []
    affiliations = []
    name_identifiers = []
    roles = []
    name_types = []
    name_identifier_schemes = []
    scheme_uri = []

    complex_people.each do |person|
      full_name = []
      if person.fetch(:last_name, '').present?
        full_name << person.fetch(:last_name, '').strip
        last_names << person.fetch(:last_name, '').strip
      else
        full_name << ""
        last_names << ""
      end
      if person.fetch(:first_name, '').present?
        full_name << person.fetch(:first_name, '').strip
        first_names << person.fetch(:first_name, '').strip
      else
        first_names << ""
      end

      names << full_name.join(', ') if full_name.any?
      name_types << "Personal"
      if person.fetch(:affiliation, '').present?
        affiliations << person.fetch(:affiliation, '').strip
      else
        affiliations << ""
      end

      if person.fetch(:orcid, '').present?
        name_identifiers << person.fetch(:orcid, '').strip
        name_identifier_schemes << "Orcid"
        scheme_uri << "https://orcid.org"
      else
        name_identifiers << ""
        name_identifier_schemes << ""
        scheme_uri << ""
      end

      if person.fetch(:role, '').present?
        roles << person.fetch(:role, '').strip
      else
        roles << ""
      end
    end

    people = {
      names: names.join("; "),
      first_names: first_names.join("; "),
      last_names: last_names.join("; "),
      affiliations: affiliations.join("; "),
      name_identifiers: name_identifiers.join("; "),
      roles: roles.join("; "),
      name_types: name_types.join("; "),
      name_identifier_schemes: name_identifier_schemes.join("; "),
      scheme_uri: scheme_uri.join("; ")
    }
    people
  end

  def data_from_complex_funding_reference(complex_funding_reference)
    funder_name = []
    funder_identifier = []

    complex_funding_reference.each do |funder|
      if funder.fetch(:funder_identifier, '').present?
        funder_identifier << funder.fetch(:funder_identifier, '').strip
      end

      if funder.fetch(:funder_name, '').present?
        funder_name << funder.fetch(:funder_name, '').strip
      end
    end

    @meta["DataCite-funderIdentifier"] = funder_identifier.join("; ")
    @meta["DataCite-funderName"] = funder_name.join("; ")
    @meta["DataCite-FundingReference"] = "DFG"
    @meta["DataCite-funderIdentifierType"] = "Crossref Funder ID"
  end

  def data_from_complex_date(complex_dates)
    @meta["Record date"] = ""
    @meta["DataCite-Date"] = ""
    @meta["DataCite-dateType"] = ""
    @meta["DublinCore-Date"] = ""
    @meta["DataCite-PublicationYear"] = ""
    date = []
    record_date = []
    date_type = []

    complex_dates.each do |complex_date|
      date_description = complex_date.fetch(:description, '')
      date_value = complex_date.fetch(:description, '')
      if date_description.present? and date.present?
        term = DateService.new.find_by_id(date_description)

        if term["label"] == "Recorded"
          record_date << date_value
        elsif term["label"] == "Created"
          date << date_value
          date_type << 'Created'
        elsif term["label"] == "Published" 
          @meta["DataCite-PublicationYear"] = date_value.to_date.year.to_s
        end
      end
    end

    @meta["Record date"] = record_date.join(';')
    @meta["DataCite-Date"] = date.join(';')
    @meta["DataCite-dateType"] = date_type.join(';')
    @meta["DublinCore-Date"] = date.join(';')
  end

  def data_from_complex_relation(complex_relations)
    @meta["DublinCore-Relation"] = ""
    @meta["DataCite-RelatedIdentifier"] = ""
    @meta["DataCite-relatedIdentifierType"] = ""
    @meta["DataCite-relationType"] = ""
    urls = []
    relations = []

    complex_relations.each do |complex_relation|
      urls << complex_relation.fetch(:url, '').strip if complex_relation.fetch(:url, '').present?
      relations << complex_relation.fetch(:relationship, '').strip if complex_relation.fetch(:relationship, '').present?
    end

    relation_urls = urls.join("; ")

    @meta["DublinCore-Relation"] = relation_urls
    @meta["DataCite-RelatedIdentifier"] = relation_urls
    @meta["DataCite-relationType"] = relations.join("; ") if relations.any?
    @meta["DataCite-relatedIdentifierType"] = relation_urls.downcase.include?('doi') ? "DOI" : "URL"
  end

end
